@*******************************
@* GBA Graphics Library        *
@*  by Jeff Frohwein           *
@* http://www.devrs.com        *
@*******************************
@ v1.0223 - Original release

@ Solid fill an area in GBA graphics modes 3/5 in ARM asm
@ Entry: x = start X coordinate
@        y = start Y coordinate
@       fc = fill color
@       bc = boundary color
@     addr = screen base addr
@ (void) gfxFill (u8 x, u8 y, u16 fc, u16 bc, u32 addr);

        .ARM
        .ALIGN
        .GLOBL  gfxFill
@ r0 = x
@ r1 = y
@ r2 = fill color
@ r3 = boundary color
@ r4 = pixel addr
@ r5 = stack count
@ r6 = tmp test pixel

gfxFill:
        sub     sp,sp,#4         @ protect the stack so addr doesn't
                                 @  doesn't get overwritten.
        stmfd   r13!,{r0-r6,r14}

        ldr     r6,[r13,#36]     @ Get the screen base address

@ Calculate start address
        mov     r4,#480
        mul     r5,r4,r1
        add     r5,r5,r0,lsl #1
        add     r4,r5,r6         @ r4 = start address

        mov     r5,#0            @ set points on stack count to 0

_SF1:   bl      _SFPntTest       @ Is point filled?
        bcs     _SF4             @ yes

_SF2:   sub     r1,r1,#1
        sub     r4,r4,#480
        bl      _SFPntTest       @ Is point above this one set?
        bcc     _SF2             @ No, so check again
        add     r1,r1,#1
        add     r4,r4,#480

_SF3:   sub     r0,r0,#1
        sub     r4,r4,#2
        bl      _SFPntTest       @ Is point on the left set ?
        bcc     _SF2             @ No, so check above again
        add     r0,r0,#1
        add     r4,r4,#2

@ Plot a pixel at r0,r1 of color r2

        strh    r2,[r4]

        add     r1,r1,#1
        add     r4,r4,#480
        bl      _SFPntTest       @ Is point below set ?
        bcc     _SF5             @ No

        sub     r1,r1,#1
        add     r0,r0,#1
        sub     r4,r4,#480
        add     r4,r4,#2
        bl      _SFPntTest       @ Is point on the right set ?
        bcc     _SF2             @ No
        sub     r0,r0,#1
        sub     r4,r4,#2

@ Are there any more points on stack?

_SF4:   cmp     r5,#0
        beq     _SF8             @ No

@ Get point on stack

        ldmfd   r13!,{r0,r1}

        sub     r5,r5,#1
        b       _SF1

_SF5:
        sub     r1,r1,#1
        add     r0,r0,#1
        sub     r4,r4,#480
        add     r4,r4,#2
        bl      _SFPntTest       @ Is point on the right set ?
        bcs     _SF6             @ Yes

        add     r1,r1,#1
        add     r4,r4,#480
        bl      _SFPntTest       @ Is lower right point set ?
        bcc     _SF7             @ No

        sub     r1,r1,#1
        sub     r4,r4,#480

@ Save point position on stack

        add     r5,r5,#1         @ Increment points on stack count
@@@@ ADD OVERFLOW TEST HERE @@@@@

        stmfd   r13!,{r0,r1}     @ Save point on stack

@ Point below is not set

_SF6:   add     r1,r1,#1
        sub     r0,r0,#1
        add     r4,r4,#480
        sub     r4,r4,#2
        b       _SF3

_SF7:   sub     r0,r0,#1
        sub     r4,r4,#2
        b       _SF3

@ Exit fill routine

_SF8:
        ldmfd   r13!,{r0-r6,r14}
        add     sp,sp,#4
        bx      r14

@ Clear Carry if pixel r0,r1 needs to be filled

_SFPntTest:
        cmps    r0,#240         @ is x coord out of bounds?
        bxcs    lr              @ yes, exit with carry set
        cmps    r1,#160         @ is y coord out of founds?
        bxcs    lr              @ yes exit with carry set

        ldrh    r6,[r4]

        cmps    r6,r2           @ Have we hit a filled spot?
        bxeq    lr              @ yes, exit with carry set

        cmps    r6,r3           @ Have we hit the boundary?
        bxeq    lr              @ yes, exit with carry set

@ Clear carry flag
        movs    r6,r0,lsl #1
        bx      lr


